/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "condor_common.h"
#include "MyString.h"
#include "which.h"
#include "string_list.h"
#include "condor_distribution.h"
#include "condor_config.h"
#include "env.h"
#include "dagman_multi_dag.h"
#include "basename.h"
#include "read_multiple_logs.h"
#include "condor_getcwd.h"


#ifdef WIN32
const char* dagman_exe = "condor_dagman.exe";
#else
const char* dagman_exe = "condor_dagman";
#endif

struct SubmitDagOptions
{
	// these options come from the command line
	bool bSubmit;
	bool bVerbose;
	bool bForce;
	MyString strNotification;
	MyString strJobLog;
	MyString strStorkLog;
	int iMaxIdle;
	int iMaxJobs;
	int iMaxPre;
	int iMaxPost;
	MyString strRemoteSchedd;
	bool bNoEventChecks;
	bool bAllowLogError;
	int iDebugLevel;
	MyString primaryDagFile;
	StringList	dagFiles;
	MyString strDagmanPath;
	bool useDagDir;
	MyString strDebugDir;
	MyString strConfigFile;
	
	// non-command line options
	MyString strLibOut;
	MyString strLibErr;
	MyString strDebugLog;
	MyString strSchedLog;
	MyString strSubFile;
	MyString strRescueFile;
	MyString strLockFile;

	SubmitDagOptions() 
	{ 
		bSubmit = true;
		bVerbose = false;
		bForce = false;
		strNotification = "";
		strJobLog = "";
		iMaxIdle = 0;
		iMaxJobs = 0;
		iMaxPre = 0;
		iMaxPost = 0;
		strRemoteSchedd = "";
		bNoEventChecks = false;
		bAllowLogError = false;
		iDebugLevel = 3;
		primaryDagFile = "";
		useDagDir = false;
		strConfigFile = "";
	}

};

int printUsage(); // NOTE: printUsage calls exit(1), so it doesnt return
void parseCommandLine(SubmitDagOptions &opts, int argc, char *argv[]);
int submitDag( SubmitDagOptions &opts );

int main(int argc, char *argv[])
{
	printf("\n");

		// Set up the dprintf stuff to write to stderr, so that Condor
		// libraries which use it will write to the right place...
	Termlog = true;
	dprintf_config("condor_submit_dag"); 
	DebugFlags = D_ALWAYS;
	config();

	SubmitDagOptions opts;
	myDistro->Init( argc, argv );
	parseCommandLine(opts, argc, argv);
	
	opts.strLibOut = opts.primaryDagFile + ".lib.out";
	opts.strLibErr = opts.primaryDagFile + ".lib.err";

	if ( opts.strDebugDir != "" ) {
		opts.strDebugLog = opts.strDebugDir + DIR_DELIM_STRING +
					condor_basename( opts.primaryDagFile.Value() );
	} else {
		opts.strDebugLog = opts.primaryDagFile;
	}
	opts.strDebugLog += ".dagman.out";

	opts.strSchedLog = opts.primaryDagFile + ".dagman.log";
	opts.strSubFile = opts.primaryDagFile + ".condor.sub";

	MyString	rescueDagBase;

		// If we're running each DAG in its own directory, write any rescue
		// DAG to the current directory, to avoid confusion (since the
		// rescue DAG must be run from the current directory).
	if ( opts.useDagDir ) {
		if ( !condor_getcwd( rescueDagBase ) ) {
			fprintf( stderr, "ERROR: unable to get cwd: %d, %s\n",
					errno, strerror(errno) );
			return 1;
		}
		rescueDagBase += DIR_DELIM_STRING;
		rescueDagBase += condor_basename(opts.primaryDagFile.Value());
	} else {
		rescueDagBase = opts.primaryDagFile;
	}

		// If we're running multiple DAGs, put "_multi" in the rescue
		// DAG name to indicate that the rescue DAG is for *all* of
		// the DAGs we're running.
	if ( opts.dagFiles.number() > 1 ) {
		rescueDagBase += "_multi";
	}

	opts.strRescueFile = rescueDagBase + ".rescue";

	opts.strLockFile = opts.primaryDagFile + ".lock";

	return submitDag( opts );
}

// utility fcns for submitDag
void ensureOutputFilesExist(const SubmitDagOptions &opts);
void writeSubmitFile(/* const */ SubmitDagOptions &opts);

int
submitDag( SubmitDagOptions &opts )
{
	if (opts.strDagmanPath == "" ) {
		opts.strDagmanPath = which( dagman_exe );
	}

	if (opts.strDagmanPath == "")
	{
		fprintf( stderr, "ERROR: can't find %s in PATH, aborting.\n",
				 dagman_exe );
		return 1;
	}
		
	ensureOutputFilesExist(opts);

	if (opts.strJobLog != "")
	{
		printf("Assuming %s is the log file shared by all jobs in this DAG.\n", opts.strJobLog.Value());
		printf("Please be sure you know what you're doing and that all jobs use this file.\n");
	}
	else
	{
		printf("Checking all your submit files for log file names.\n");
		printf("This might take a while... \n");

		StringList	condorLogFiles;
		StringList	storkLogFiles;

		MyString	msg;
		if ( !GetLogFiles( opts.dagFiles, opts.useDagDir, condorLogFiles,
					storkLogFiles, msg ) ) {
			fprintf( stderr, "ERROR: %s\n", msg.Value() );
			if ( opts.bAllowLogError ) {
				fprintf( stderr, "Continuing anyhow because of "
							"-AllowLogError flag (beware!)\n" );
			} else {
				fprintf( stderr, "Aborting -- try again with the "
							"-AllowLogError flag if you *really* think "
							"this shouldn't be a fatal error\n" );
				return 1;
			}
		}

		if (LogFileNfsError(condorLogFiles, storkLogFiles)) {
			return 1;
		}

		condorLogFiles.rewind();
		storkLogFiles.rewind();

		if ( condorLogFiles.number() > 0 ) {
			opts.strJobLog = condorLogFiles.next();
		} else if ( storkLogFiles.number() > 0 ) {
			opts.strJobLog = storkLogFiles.next();
		}
		printf("Done.\n");
	}

	MyString	msg;
	if ( !GetConfigFile( opts.dagFiles, opts.useDagDir,
				opts.strConfigFile, msg) ) {
		fprintf( stderr, "ERROR: %s\n", msg.Value() );
		return 1;
	}

	writeSubmitFile(opts);

	printf("-----------------------------------------------------------------------\n");
	printf("File for submitting this DAG to Condor           : %s\n", 
			opts.strSubFile.Value());
	printf("Log of DAGMan debugging messages                 : %s\n",
		   	opts.strDebugLog.Value());
	printf("Log of Condor library output                     : %s\n", 
			opts.strLibOut.Value());
	printf("Log of Condor library error messages             : %s\n", 
			opts.strLibErr.Value());
	printf("Log of the life of condor_dagman itself          : %s\n",
		   	opts.strSchedLog.Value());
	printf("\n");
	printf("Condor Log file for all jobs of this DAG         : %s\n", 
			opts.strJobLog.Value());

	if (opts.strStorkLog != "") {
		printf("Stork Log file for all DaP jobs of this DAG      : %s\n",
			   	opts.strStorkLog.Value());
	}

	if (opts.bSubmit)
	{
		MyString strCmdLine = "condor_submit " + opts.strRemoteSchedd + " " + opts.strSubFile;
		int retval = system(strCmdLine.Value());
		if( retval != 0 ) {
			fprintf( stderr, "ERROR: condor_submit failed; aborting.\n" );
			return 1;
		}
	}
	else
	{
		printf("-no_submit given, not submitting DAG to Condor.  You can do this with:\n");
		printf("\"condor_submit %s\"\n", opts.strSubFile.Value());
	}
	printf("-----------------------------------------------------------------------\n");

	return 0;
}

MyString makeString(int iValue)
{
	char psToReturn[16];
	sprintf(psToReturn,"%d",iValue);
	return MyString(psToReturn);
}

bool fileExists(const MyString &strFile)
{
	int fd = safe_open_wrapper(strFile.Value(), O_RDONLY);
	if (fd == -1)
		return false;
	close(fd);
	return true;
}

void ensureOutputFilesExist(const SubmitDagOptions &opts)
{
	if (opts.bForce)
	{
		unlink(opts.strSubFile.Value());
		unlink(opts.strSchedLog.Value());
		unlink(opts.strLibOut.Value());
		unlink(opts.strLibErr.Value());
		unlink(opts.strRescueFile.Value());
	}

	bool bHadError = false;
	if (fileExists(opts.strSubFile))
	{
		fprintf( stderr, "ERROR: \"%s\" already exists.\n",
				 opts.strSubFile.Value() );
		bHadError = true;
	}
	if (fileExists(opts.strLibOut))
	{
		fprintf( stderr, "ERROR: \"%s\" already exists.\n",
				 opts.strLibOut.Value() );
		bHadError = true;
	}
	if (fileExists(opts.strLibErr))
	{
		fprintf( stderr, "ERROR: \"%s\" already exists.\n",
				 opts.strLibErr.Value() );
		bHadError = true;
	}
	if (fileExists(opts.strSchedLog))
	{
		fprintf( stderr, "ERROR: \"%s\" already exists.\n",
				 opts.strSchedLog.Value() );
		bHadError = true;
	}
	if (fileExists(opts.strRescueFile))
	{
		fprintf( stderr, "ERROR: \"%s\" already exists.\n",
				 opts.strRescueFile.Value() );
	    fprintf( stderr, "  You may want to resubmit your DAG using that "
				 "file, instead of \"%s\"\n", opts.primaryDagFile.Value());
	    fprintf( stderr, "  Look at the Condor manual for details about DAG "
				 "rescue files.\n" );
	    fprintf( stderr, "  Please investigate and either remove \"%s\",\n",
				 opts.strRescueFile.Value() );
	    fprintf( stderr, "  or use it as the input to condor_submit_dag.\n" );
		bHadError = true;
	}

	if (bHadError) 
	{
	    fprintf( stderr, "\nSome file(s) needed by %s already exist.  ",
				 dagman_exe );
	    fprintf( stderr, "Either rename them,\nor use the \"-f\" option to "
				 "force them to be overwritten.\n" );
	    exit( 1 );
	}
}

void writeSubmitFile(/* const */ SubmitDagOptions &opts)
{
	FILE *pSubFile = safe_fopen_wrapper(opts.strSubFile.Value(), "w");
	if (!pSubFile)
	{
		fprintf( stderr, "ERROR: unable to create submit file %s\n",
				 opts.strSubFile.Value() );
		exit( 1 );
	}

    fprintf(pSubFile, "# Filename: %s\n", opts.strSubFile.Value());

    fprintf(pSubFile, "# Generated by condor_submit_dag ");
	opts.dagFiles.rewind();
	char *dagFile;
	while ( (dagFile = opts.dagFiles.next()) != NULL ) {
    	fprintf(pSubFile, "%s ", dagFile);
	}
    fprintf(pSubFile, "\n");

    fprintf(pSubFile, "universe\t= scheduler\n");
    fprintf(pSubFile, "executable\t= %s\n", opts.strDagmanPath.Value());
	fprintf(pSubFile, "getenv\t\t= True\n");
	fprintf(pSubFile, "output\t\t= %s\n", opts.strLibOut.Value());
    fprintf(pSubFile, "error\t\t= %s\n", opts.strLibErr.Value());
    fprintf(pSubFile, "log\t\t= %s\n", opts.strSchedLog.Value());
#if !defined ( WIN32 )
    fprintf(pSubFile, "remove_kill_sig\t= SIGUSR1\n" );
#endif

		// ensure DAGMan is automatically requeued by the schedd if it
		// exits abnormally or is killed (e.g., during a reboot)
	const char *defaultRemoveExpr = "( ExitSignal =?= 11 || "
				"(ExitCode =!= UNDEFINED && ExitCode >=0 && ExitCode <= 2))";
	MyString removeExpr(defaultRemoveExpr);
	char *tmpRemoveExpr = param("DAGMAN_ON_EXIT_REMOVE");
	if ( tmpRemoveExpr ) {
		removeExpr = tmpRemoveExpr;
		free(tmpRemoveExpr);
	}
    fprintf(pSubFile, "# Note: default on_exit_remove expression:\n");
	fprintf(pSubFile, "# %s\n", defaultRemoveExpr);
	fprintf(pSubFile, "# attempts to ensure that DAGMan is automatically\n");
	fprintf(pSubFile, "# requeued by the schedd if it exits abnormally or\n");
    fprintf(pSubFile, "# is killed (e.g., during a reboot).\n");
    fprintf(pSubFile, "on_exit_remove\t= %s\n", removeExpr.Value() );

    fprintf(pSubFile, "copy_to_spool\t= False\n" );

	ArgList args;

	args.AppendArg("-f");
	args.AppendArg("-l");
	args.AppendArg(".");
	args.AppendArg("-Debug");
	args.AppendArg(opts.iDebugLevel);
	args.AppendArg("-Lockfile");
	args.AppendArg(opts.strLockFile.Value());

	if ( opts.strJobLog == "" ) {
			// Condor_dagman can't parse command-line args if we have
			// an empty value for the log file name.
		opts.strJobLog = opts.primaryDagFile + ".log";
	}
	args.AppendArg("-Condorlog");
	args.AppendArg(opts.strJobLog.Value());
	
	opts.dagFiles.rewind();
	while ( (dagFile = opts.dagFiles.next()) != NULL ) {
		args.AppendArg("-Dag");
		args.AppendArg(dagFile);
	}

	args.AppendArg("-Rescue");
	args.AppendArg(opts.strRescueFile.Value());
    if(opts.iMaxIdle) 
	{
		args.AppendArg("-MaxIdle");
		args.AppendArg(opts.iMaxIdle);
    }
    if(opts.iMaxJobs) 
	{
		args.AppendArg("-MaxJobs");
		args.AppendArg(opts.iMaxJobs);
    }
    if(opts.iMaxPre) 
	{
		args.AppendArg("-MaxPre");
		args.AppendArg(opts.iMaxPre);
    }
    if(opts.iMaxPost) 
	{
		args.AppendArg("-MaxPost");
		args.AppendArg(opts.iMaxPost);
    }
    if(opts.strStorkLog != "") 
	{
		args.AppendArg("-Storklog");
		args.AppendArg(opts.strStorkLog.Value());
    }
	if(opts.bNoEventChecks)
	{
		// strArgs += " -NoEventChecks";
		printf( "Warning: -NoEventChecks is ignored; please use "
					"the DAGMAN_ALLOW_EVENTS config parameter instead\n");
	}
	if(opts.bAllowLogError)
	{
		args.AppendArg("-AllowLogError");
	}
	if(opts.useDagDir)
	{
		args.AppendArg("-UseDagDir");
	}

	MyString arg_str,args_error;
	if(!args.GetArgsStringV1WackedOrV2Quoted(&arg_str,&args_error)) {
		fprintf(stderr,"Failed to insert arguments: %s",args_error.Value());
		exit(1);
	}
    fprintf(pSubFile, "arguments\t= %s\n", arg_str.Value());

	Env env;
	env.SetEnv("_CONDOR_DAGMAN_LOG",opts.strDebugLog.Value());
	env.SetEnv("_CONDOR_MAX_DAGMAN_LOG=0");
	if ( opts.strConfigFile != "" ) {
		if ( access( opts.strConfigFile.Value(), F_OK ) != 0 ) {
			fprintf( stderr, "ERROR: unable to read config file %s "
						"(error %d, %s)\n",
						opts.strConfigFile.Value(), errno, strerror(errno) );
			exit(1);
		}
		env.SetEnv("_CONDOR_DAGMAN_CONFIG_FILE", opts.strConfigFile.Value());
	}

	MyString env_str;
	MyString env_errors;
	if(!env.getDelimitedStringV1RawOrV2Quoted(&env_str,&env_errors)) {
		fprintf(stderr,"Failed to insert environment: %s",env_errors.Value());
		exit(1);
	}
    fprintf(pSubFile, "environment\t= %s\n",env_str.Value());

    if(opts.strNotification != "") 
	{	
		fprintf(pSubFile, "notification\t= %s\n", opts.strNotification.Value());
    }
    fprintf(pSubFile, "queue\n");

	fclose(pSubFile);
	
}

void parseCommandLine(SubmitDagOptions &opts, int argc, char *argv[])
{
	for (int iArg = 1; iArg < argc; iArg++)
	{
		MyString strArg = argv[iArg];

		if (strArg[0] != '-')
		{
				// We assume an argument without a leading hyphen is
				// a DAG file name.
			opts.dagFiles.append(strArg.Value());
			if ( opts.primaryDagFile == "" ) {
				opts.primaryDagFile = strArg;
			}
		}
		else if (opts.primaryDagFile != "")
		{
				// Disallow hyphen args after DAG file name(s).
			printf("ERROR: no arguments allowed after DAG file name(s)\n");
			printUsage();
		}
		else
		{
			strArg.lower_case();

			// Note: in checking the argument names here, we only check for
			// as much of the full name as we need to unambiguously define
			// the argument.
			if (strArg.find("-no_s") != -1) // -no_submit
			{
				opts.bSubmit = false;
			}
			else if (strArg.find("-v") != -1) // -verbose
			{
				opts.bVerbose = true;
			}
			else if (strArg.find("-f") != -1) // -force
			{
				opts.bForce = true;
			}
			else if (strArg.find("-not") != -1) // -notification
			{
				opts.strNotification = argv[++iArg];
			}
			else if (strArg.find("-l") != -1) // -log
			{
				opts.strJobLog = argv[++iArg];
			}
			else if (strArg.find("-maxi") != -1) // -maxidle
			{
				opts.iMaxIdle = atoi(argv[++iArg]);
			}
			else if (strArg.find("-maxj") != -1) // -maxjobs
			{
				opts.iMaxJobs = atoi(argv[++iArg]);
			}
			else if (strArg.find("-maxpr") != -1) // -maxpre
			{
				opts.iMaxPre = atoi(argv[++iArg]);
			}
			else if (strArg.find("-maxpo") != -1) // -maxpost
			{
				opts.iMaxPost = atoi(argv[++iArg]);
			}
			else if (strArg.find("-r") != -1) // submit to remote schedd
			{
				opts.strRemoteSchedd = MyString("-r ") + argv[++iArg];
			}
			else if (strArg.find("-dagman") != -1)
			{
				opts.strDagmanPath = argv[++iArg];
			}
			else if (strArg.find("-storklog") != -1)
			{
				opts.strStorkLog = argv[++iArg];
			}
			else if (strArg.find("-d") != -1) // -debug
			{
				opts.iDebugLevel = atoi(argv[++iArg]);
			}
			else if (strArg.find("-noev") != -1) // -noeventchecks
			{
				opts.bNoEventChecks = true;
			}
			else if (strArg.find("-allow") != -1) // -allowlogerror
			{
				opts.bAllowLogError = true;
			}
			else if (strArg.find("-use") != -1) // -usedagdir
			{
				opts.useDagDir = true;
			}
			else if (strArg.find("-out") != -1) // -outfile_dir
			{
				opts.strDebugDir = argv[++iArg];
			}
			else if (strArg.find("-con") != -1) // -config
			{
				opts.strConfigFile = argv[++iArg];
					// Internally we deal with all configuration file paths
					// as full paths, to make it easier to determine whether
					// several paths point to the same file.
				MyString	errMsg;
				if (!MakePathAbsolute(opts.strConfigFile, errMsg)) {
					fprintf( stderr, "%s\n", errMsg.Value() );
	    			exit( 1 );
				}
			}
			else
			{
				fprintf( stderr, "ERROR: unknown option %s\n", strArg.Value() );
				printUsage();
			}
		}
	}

	if (opts.primaryDagFile == "")
	{
		fprintf( stderr, "ERROR: no dag file specified; aborting.\n" );
		printUsage();
	}
}

// condor_submit_dag diamond.dag
int printUsage() 
{
    printf("Usage: condor_submit_dag [options] dag_file [dag_file_2 ... dag_file_n]\n");
    printf("  where dag_file1, etc., is the name of a DAG input file\n");
    printf("  and where [options] is one or more of:\n");
	printf("    -dagman <path>      (Full path to an alternate condor_dagman executable)\n");
    printf("    -no_submit          (DAG is not submitted to Condor)\n");
    printf("    -verbose            (Verbose error messages from condor_submit_dag)\n");
    printf("    -force              (Overwrite files condor_submit_dag uses if they exist)\n");
    printf("    -r <schedd_name>    (Submit to the specified remote schedd)\n");
	printf("    -maxidle <number>   (Maximum number of idle nodes to allow)\n");
    printf("    -maxjobs <number>   (Maximum number of jobs ever submitted at once)\n");
    printf("    -MaxPre <number>    (Maximum number of PRE scripts to run at once)\n");
    printf("    -MaxPost <number>   (Maximum number of POST scripts to run at once)\n");
    printf("    -log <filename>     (Deprecated -- don't use)\n");
// -->STORK
    printf("    -storklog <filename> (Specify the Stork log file shared by all DaP jobs\n");
    printf("        in the DAG)\n");
// <--STORK
    printf("    -notification <value> (Determines how much email you get from Condor.\n");
    printf("        See the condor_submit man page for values.)\n");
    printf("    -NoEventChecks      (Now ignored -- use DAGMAN_ALLOW_EVENTS)\n"); 
    printf("    -AllowLogError      (Allows the DAG to attempt execution even if the log\n");
    printf("        reading code finds errors when parsing the submit files)\n"); 
	printf("    -UseDagDir          (Run DAGs in directories specified in DAG file paths)\n");
    printf("    -debug <number>     (Determines how verbosely DAGMan logs its work\n");
    printf("         about the life of the condor_dagman job.  'value' must be\n");
    printf("         an integer with a value of 0-7 inclusive.)\n");
    printf("    -outfile_dir <path> (Directory into which to put the dagman.out file,\n");
	printf("         instead of the default\n");
	exit(1);
}
