/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#define _CONDOR_ALLOW_OPEN 1
#include "condor_common.h"
#include "condor_nfs.h"
#include "condor_classad.h"
#include "condor_debug.h"
#include "condor_io.h"
#include "condor_sys.h"
#include "condor_mmap.h"
#include "startup.h"
#include "pseudo_ops.h"
#include "../condor_shadow.V6/job_report.h"

	/*
	Have assert() really call EXCEPT instead of doing a sigabrt 
	and dumping core.  This change results in a shadow exception when
	communication with the remote job is interrupted in the middle of 
	a remote system call, instead of the shadow dumping core in the
	user's current working directory.
	Note: the definition to CONDOR_ASSERT() macro is in condor_debug.h
	-Todd <tannenba@cs.wisc.edu> 10/10/2000
	*/
#ifdef assert
#	undef assert
#endif
#define assert(cond) CONDOR_ASSERT(cond)

/* get rid of the access remap to access_euid for the syscall library */
#ifdef access
#	undef access
#endif

#if defined(Solaris)
extern "C" {
	int flock( int, int );
}
#endif

#if defined(OSF1) 
extern "C" {
# if defined DUX4
	int statfs( char*, struct statfs*, int );
	int fstatfs ( int fd, struct statfs *buf, int len );
# elif defined DUX5
        int _F64_statfs( char*, struct statfs*, ... );
        int fstatfs ( int fd, struct statfs *buf, ... );
# endif
	int flock( int, int );
	int getdirentries ( int fd, struct dirent *buf, size_t nbytes, off_t *basep );
	int getfsstat ( struct statfs buf[], long bufsize, int flags );
	int table ( int id, int index, char *addr, int nel, unsigned int lel );
	int getdomainname ( char *name, int namelen );
	int getaddressconf ( struct addressconf *buf, size_t length );
	int uswitch ( int cmd, int value ); 
}
#endif

#if defined(HPUX)
extern "C" {
	int flock( int, int );
}
#endif

#if defined(LINUX)
extern "C" {
	
	ssize_t getdirentries(int fd, char *buf, size_t nbytes, off_t *basep);
	ssize_t getdirentries64(int fd, char *buf, size_t nbytes, off64_t *basep);

	/*
	There is no public interface to getdents in the Linux C library.
	So, here we provide one based on a better-known public function.
	*/

	/* It angers me that this is here because basically it creates a definition
		for getdents when linking with the standard universe shadow.
		-psilord */

	static int getdents( int fd, struct dirent *buf, size_t nbytes ) {
		off_t basep;
		basep = lseek(fd,0,SEEK_CUR);
		return getdirentries(fd,(char*)buf,nbytes,&basep);
	}

	static int getdents64( int fd, struct dirent64 *buf, size_t nbytes ) {
		off64_t basep;
		basep = lseek64(fd,0,SEEK_CUR);
		return getdirentries64(fd,(char*)buf,nbytes,&basep);
	}
}
#endif

#if defined(Solaris) || defined(LINUX)
/* there doesn't seem to be a correct prototype of this in the header files.. */
extern "C" {
int getdents64( int fd, struct dirent64 *buf, size_t nbytes );
}
#endif

#include "../condor_syscall_lib/syscall_param_sizes.h"

extern ReliSock *syscall_sock;

#if defined(SYSCALL_DEBUG)
	extern char *SyscallLabel;
#endif

int CONDOR_NotSupported( int num, ... );
int CONDOR_Ignored( int num, ... );

#if defined(USE_STUB_CLUMPS)
int do_REMOTE_syscall1(int);
int do_REMOTE_syscall2(int);
int do_REMOTE_syscall3(int);
int do_REMOTE_syscall4(int);
int do_REMOTE_syscall5(int);
#endif

#if !defined(PVM_RECEIVE)
int
do_REMOTE_syscall()
{
	int condor_sysnum;
	int	rval;

	syscall_sock->decode();

	rval = syscall_sock->code(condor_sysnum);
	if(!rval) {
		char *error;
		if(errno==0) {
			error = "disconnected";
		} else {
			error = strerror(errno);
		}
		EXCEPT("Unable to talk to job: %s\n",error);
	}

#if defined(SYSCALL_DEBUG)
	sprintf( SyscallLabel, "%-6d", condor_sysnum );
#endif

	dprintf(D_SYSCALLS,
		"Got request for syscall %d <%s>\n",
		condor_sysnum, _condor_syscall_name(condor_sysnum)
	);

	job_report_store_call( condor_sysnum );

	switch( condor_sysnum ) {
#else

#include "joblist.h"
extern struct tagJobList	*JobList;

#include "prototypes.h"

int
do_REMOTE_syscall(ReliSock *sock, int process)
{
	int condor_sysnum;
	int rval;
	int	read_ok;

	syscall_sock->decode();

	read_ok = syscall_sock->code(condor_sysnum);

	if( !read_ok ) {
		dprintf( D_ALWAYS, 
		"Failed to get syscall_code for proc %d removing..\n", process);

		RemoveProc( process );
		return -1;
	}

	dprintf( D_SYSCALLS, "Shadow: got request for syscall %d <%s>\n",
			condor_sysnum, _condor_syscall_name(condor_sysnum) );

	switch( condor_sysnum ) {
        case CONDOR_work_request:
          {
                ClassAd *ad = NULL;
                char *a_out;
                char *targ;
                char *orig;
                condor_signal_t *kill_sig;
                condor_errno_t terrno;

                kill_sig = (condor_signal_t *)malloc( (unsigned)INT_SIZE );
                orig = NULL;
                targ = NULL;
                a_out = NULL;
		/* space for ad allocated by pseudo_work_req */
                assert( syscall_sock->end_of_message() );;

                errno = 0;
                rval = pseudo_work_request( ad, a_out, targ, orig, (int *)kill_sig );
                terrno = (condor_errno_t)errno;
                dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno);

                syscall_sock->encode();
                assert( syscall_sock->code(rval) );
                if( rval < 0 ) {
                        assert( syscall_sock->code(terrno) );
                }
                if( rval >= 0 ) {
                        assert( syscall_sock->code(kill_sig) );
                        assert( syscall_sock->code(orig) );
                        assert( syscall_sock->code(targ) );
                        assert( syscall_sock->code(a_out) );
			assert( ad->put(*syscall_sock) );
                }
                free( (char *)kill_sig );
                free( (char *)orig );
                free( (char *)targ );
                free( (char *)a_out );
		if ( ad != NULL ) delete ad;
                assert( syscall_sock->end_of_message() );;
                return 0;
        }
#endif  // of ifdef PVM_RECEIVE
	case CONDOR_startup_info_request:
	  {
		STARTUP_INFO *s;
		condor_errno_t terrno;

		s = (STARTUP_INFO *)malloc( (unsigned)STARTUP_INFO_SIZE );
		memset( s, 0, (unsigned)STARTUP_INFO_SIZE );
		assert( syscall_sock->end_of_message() );;

		errno = 0;
		rval = pseudo_startup_info_request( s );
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );
		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		if( rval >= 0 ) {
			assert( syscall_sock->code(s) );
		}
		free( s->cmd );
		free( s->args_v1or2 );
		free( s->env_v1or2 );
		free( s->iwd );
		free( s );

		assert( syscall_sock->end_of_message() );;
		return 0;
		
	}
	case CONDOR_open:
	  {
		char *  path;
		open_flags_t   flags;
		long   lastarg;
		condor_errno_t terrno;

		assert( syscall_sock->code(flags) );
		dprintf( D_SYSCALLS, "	flags = %d\n", flags );
		assert( syscall_sock->code(lastarg) );
		path = NULL;
		assert( syscall_sock->code(path) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = open( path , flags , lastarg);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		free( (char *)path );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_close:
	  {
		int   fd;
		condor_errno_t terrno;

		assert( syscall_sock->code(fd) );
		dprintf( D_SYSCALLS, "	fd = %d\n", fd );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = close( fd);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_unlink:
	  {
		char *  path;
		condor_errno_t terrno;

		path = NULL;
		assert( syscall_sock->code(path) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = unlink( path);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		free( (char *)path );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_rename:
	  {
		char *  from;
		char *  to;
		condor_errno_t terrno;

		to = NULL;
		from = NULL;
		assert( syscall_sock->code(to) );
		assert( syscall_sock->code(from) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_rename( from , to);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		free( (char *)to );
		free( (char *)from );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_getwd:
	  {
		char *  path_name;
		condor_errno_t terrno;

		path_name = NULL;
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_getwd( path_name);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		if( rval >= 0 ) {
			assert( syscall_sock->code(path_name) );
		}
		free( (char *)path_name );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_geteuid:
	  {
		condor_errno_t terrno;

		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_geteuid( );
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_getegid:
	  {
		condor_errno_t terrno;

		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_getegid( );
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_register_ckpt_platform:
	  {
		char *  platform;
		int   len;
		condor_errno_t terrno;

		assert( syscall_sock->code(len) );
		dprintf( D_SYSCALLS, "	len = %d\n", len );
		platform = (char *)malloc( (unsigned)len );
		memset( platform, 0, (unsigned)len );
		assert( syscall_sock->code_bytes_bool(platform, len) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_register_ckpt_platform( platform , len);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		free( (char *)platform );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_register_fs_domain:
	  {
		char *  domain;
		condor_errno_t terrno;

		domain = NULL;
		assert( syscall_sock->code(domain) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_register_fs_domain( domain);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		free( (char *)domain );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_register_uid_domain:
	  {
		char *  domain;
		condor_errno_t terrno;

		domain = NULL;
		assert( syscall_sock->code(domain) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_register_uid_domain( domain);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		free( (char *)domain );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_get_universe:
	  {
		int *  univ_id;
		condor_errno_t terrno;

		univ_id = (int *)malloc( (unsigned)INT_SIZE );
		memset( univ_id, 0, (unsigned)INT_SIZE );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_get_universe( univ_id);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		if( rval >= 0 ) {
			assert( syscall_sock->code(univ_id) );
		}
		free( (char *)univ_id );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_get_username:
	  {
		char *  uname;
		condor_errno_t terrno;

		uname = NULL;
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_get_username( uname);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		if( rval >= 0 ) {
			assert( syscall_sock->code(uname) );
		}
		free( (char *)uname );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_extern_name:
	  {
		char *  path;
		char *  buf;
		int   bufsize;
		condor_errno_t terrno;

		assert( syscall_sock->code(bufsize) );
		dprintf( D_SYSCALLS, "	bufsize = %d\n", bufsize );
		path = NULL;
		buf = (char *)malloc( (unsigned)bufsize );
		memset( buf, 0, (unsigned)bufsize );
		assert( syscall_sock->code(path) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_extern_name( path , buf , bufsize);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		if( rval >= 0 ) {
			assert( syscall_sock->code_bytes_bool(buf, bufsize) );
		}
		free( (char *)buf );
		free( (char *)path );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_free_fs_blocks:
	  {
		char *  pathname;
		condor_errno_t terrno;

		pathname = NULL;
		assert( syscall_sock->code(pathname) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_free_fs_blocks( pathname);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		free( (char *)pathname );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_reallyexit:
	  {
		int *  status;
		struct rusage *  use_p;
		condor_errno_t terrno;

		use_p = (struct rusage *)malloc( (unsigned)RUSAGE_SIZE );
		memset( use_p, 0, (unsigned)RUSAGE_SIZE );
		status = (int *)malloc( (unsigned)INT_SIZE );
		memset( status, 0, (unsigned)INT_SIZE );
		assert( syscall_sock->code(use_p) );
		assert( syscall_sock->code(status) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_reallyexit( status , use_p);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		free( (char *)use_p );
		free( (char *)status );
		assert( syscall_sock->end_of_message() );;
		return NEG_ONE;
	}

	case CONDOR_subproc_status:
	  {
		int   id;
		int *  status;
		struct rusage *  use_p;
		condor_errno_t terrno;

		assert( syscall_sock->code(id) );
		dprintf( D_SYSCALLS, "	id = %d\n", id );
		use_p = (struct rusage *)malloc( (unsigned)RUSAGE_SIZE );
		memset( use_p, 0, (unsigned)RUSAGE_SIZE );
		status = (int *)malloc( (unsigned)INT_SIZE );
		memset( status, 0, (unsigned)INT_SIZE );
		assert( syscall_sock->code(use_p) );
		assert( syscall_sock->code(status) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_subproc_status( id , status , use_p);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		free( (char *)use_p );
		free( (char *)status );
		assert( syscall_sock->end_of_message() );;
		return NEG_ONE;
	}

	case CONDOR_image_size:
	  {
		int   kbytes;
		condor_errno_t terrno;

		assert( syscall_sock->code(kbytes) );
		dprintf( D_SYSCALLS, "	kbytes = %d\n", kbytes );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_image_size( kbytes);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_send_rusage:
	  {
		struct rusage *  use_p;
		condor_errno_t terrno;

		use_p = (struct rusage *)malloc( (unsigned)RUSAGE_SIZE );
		memset( use_p, 0, (unsigned)RUSAGE_SIZE );
		assert( syscall_sock->code(use_p) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_send_rusage( use_p);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		free( (char *)use_p );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_getrusage:
	  {
		int   who;
		struct rusage *  use_p;
		condor_errno_t terrno;

		assert( syscall_sock->code(who) );
		dprintf( D_SYSCALLS, "	who = %d\n", who );
		use_p = (struct rusage *)malloc( (unsigned)RUSAGE_SIZE );
		memset( use_p, 0, (unsigned)RUSAGE_SIZE );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_getrusage( who , use_p);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		if( rval >= 0 ) {
			assert( syscall_sock->code(use_p) );
		}
		free( (char *)use_p );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_report_error:
	  {
		char *  msg;
		condor_errno_t terrno;

		msg = NULL;
		assert( syscall_sock->code(msg) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_report_error( msg);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		free( (char *)msg );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_report_file_info:
	  {
		char *  kind;
		char *  name;
		int   open_count;
		int   read_count;
		int   write_count;
		int   seek_count;
		int   read_bytes;
		int   write_bytes;
		condor_errno_t terrno;

		assert( syscall_sock->code(open_count) );
		dprintf( D_SYSCALLS, "	open_count = %d\n", open_count );
		assert( syscall_sock->code(read_count) );
		dprintf( D_SYSCALLS, "	read_count = %d\n", read_count );
		assert( syscall_sock->code(write_count) );
		dprintf( D_SYSCALLS, "	write_count = %d\n", write_count );
		assert( syscall_sock->code(seek_count) );
		dprintf( D_SYSCALLS, "	seek_count = %d\n", seek_count );
		assert( syscall_sock->code(read_bytes) );
		dprintf( D_SYSCALLS, "	read_bytes = %d\n", read_bytes );
		assert( syscall_sock->code(write_bytes) );
		dprintf( D_SYSCALLS, "	write_bytes = %d\n", write_bytes );
		name = NULL;
		kind = NULL;
		assert( syscall_sock->code(name) );
		assert( syscall_sock->code(kind) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_report_file_info( kind , name , open_count , read_count , write_count , seek_count , read_bytes , write_bytes);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		free( (char *)name );
		free( (char *)kind );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_report_file_info_new:
	  {
		char *  name;
		int64_t   open_count;
		int64_t   read_count;
		int64_t   write_count;
		int64_t   seek_count;
		int64_t   read_bytes;
		int64_t   write_bytes;
		condor_errno_t terrno;

		assert( syscall_sock->code(open_count) );
		assert( syscall_sock->code(read_count) );
		assert( syscall_sock->code(write_count) );
		assert( syscall_sock->code(seek_count) );
		assert( syscall_sock->code(read_bytes) );
		assert( syscall_sock->code(write_bytes) );
		name = NULL;
		assert( syscall_sock->code(name) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_report_file_info_new( name , open_count , read_count , write_count , seek_count , read_bytes , write_bytes);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		free( (char *)name );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_get_std_file_info:
	  {
		int   fd;
		char *  logical_name;
		condor_errno_t terrno;

		assert( syscall_sock->code(fd) );
		dprintf( D_SYSCALLS, "	fd = %d\n", fd );
		logical_name = NULL;
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_get_std_file_info( fd , logical_name);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		if( rval >= 0 ) {
			assert( syscall_sock->code(logical_name) );
		}
		free( (char *)logical_name );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_get_file_info_new:
	  {
		char *  logical_name;
		char *  actual_url;
		condor_errno_t terrno;

		actual_url = NULL;
		logical_name = NULL;
		assert( syscall_sock->code(logical_name) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_get_file_info_new( logical_name , actual_url);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		if( rval >= 0 ) {
			assert( syscall_sock->code(actual_url) );
		}
		free( (char *)actual_url );
		free( (char *)logical_name );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_get_file_info:
	  {
		char *  logical_name;
		char *  actual_url;
		condor_errno_t terrno;

		actual_url = NULL;
		logical_name = NULL;
		assert( syscall_sock->code(logical_name) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_get_file_info( logical_name , actual_url);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		if( rval >= 0 ) {
			assert( syscall_sock->code(actual_url) );
		}
		free( (char *)actual_url );
		free( (char *)logical_name );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_std_file_info:
	  {
		int   fd;
		char *  logical_name;
		int *  pipe_fd;
		condor_errno_t terrno;

		assert( syscall_sock->code(fd) );
		dprintf( D_SYSCALLS, "	fd = %d\n", fd );
		logical_name = NULL;
		pipe_fd = (int *)malloc( (unsigned)INT_SIZE );
		memset( pipe_fd, 0, (unsigned)INT_SIZE );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_std_file_info( fd , logical_name , pipe_fd);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		if( rval >= 0 ) {
			assert( syscall_sock->code(pipe_fd) );
			assert( syscall_sock->code(logical_name) );
		}
		free( (char *)pipe_fd );
		free( (char *)logical_name );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_file_info:
	  {
		char *  logical_name;
		int *  fd;
		char *  physical_name;
		condor_errno_t terrno;

		physical_name = NULL;
		logical_name = NULL;
		fd = (int *)malloc( (unsigned)INT_SIZE );
		memset( fd, 0, (unsigned)INT_SIZE );
		assert( syscall_sock->code(logical_name) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_file_info( logical_name , fd , physical_name);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		if( rval >= 0 ) {
			assert( syscall_sock->code(fd) );
			assert( syscall_sock->code(physical_name) );
		}
		free( (char *)fd );
		free( (char *)physical_name );
		free( (char *)logical_name );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_get_buffer_info:
	  {
		int *  bytes;
		int *  block_size;
		int *  prefetch_size;
		condor_errno_t terrno;

		prefetch_size = (int *)malloc( (unsigned)INT_SIZE );
		memset( prefetch_size, 0, (unsigned)INT_SIZE );
		block_size = (int *)malloc( (unsigned)INT_SIZE );
		memset( block_size, 0, (unsigned)INT_SIZE );
		bytes = (int *)malloc( (unsigned)INT_SIZE );
		memset( bytes, 0, (unsigned)INT_SIZE );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_get_buffer_info( bytes , block_size , prefetch_size);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		if( rval >= 0 ) {
			assert( syscall_sock->code(prefetch_size) );
			assert( syscall_sock->code(block_size) );
			assert( syscall_sock->code(bytes) );
		}
		free( (char *)prefetch_size );
		free( (char *)block_size );
		free( (char *)bytes );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_get_job_attr:
	  {
		char *  attrname;
		char *  expr;
		condor_errno_t terrno;

		expr = NULL;
		attrname = NULL;
		assert( syscall_sock->code(attrname) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_get_job_attr( attrname , expr);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		if( rval >= 0 ) {
			assert( syscall_sock->code(expr) );
		}
		free( (char *)expr );
		free( (char *)attrname );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_set_job_attr:
	  {
		char *  attrname;
		char *  expr;
		condor_errno_t terrno;

		expr = NULL;
		attrname = NULL;
		assert( syscall_sock->code(expr) );
		assert( syscall_sock->code(attrname) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_set_job_attr( attrname , expr);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		free( (char *)expr );
		free( (char *)attrname );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_constrain:
	  {
		char *  expr;
		condor_errno_t terrno;

		expr = NULL;
		assert( syscall_sock->code(expr) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_constrain( expr);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		free( (char *)expr );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_get_iwd:
	  {
		char *  path;
		condor_errno_t terrno;

		path = NULL;
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_get_iwd( path);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		if( rval >= 0 ) {
			assert( syscall_sock->code(path) );
		}
		free( (char *)path );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_get_ckpt_name:
	  {
		char *  path;
		condor_errno_t terrno;

		path = NULL;
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_get_ckpt_name( path);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		if( rval >= 0 ) {
			assert( syscall_sock->code(path) );
		}
		free( (char *)path );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_get_ckpt_mode:
	  {
		condor_signal_t   sig;
		condor_errno_t terrno;

		assert( syscall_sock->code(sig) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_get_ckpt_mode( sig);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_get_ckpt_speed:
	  {
		condor_errno_t terrno;

		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_get_ckpt_speed( );
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_get_a_out_name:
	  {
		char *  path;
		condor_errno_t terrno;

		path = NULL;
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_get_a_out_name( path);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		if( rval >= 0 ) {
			assert( syscall_sock->code(path) );
		}
		free( (char *)path );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_pvm_info:
	  {
		char *  response;
		int   mytid;
		condor_errno_t terrno;

		assert( syscall_sock->code(mytid) );
		dprintf( D_SYSCALLS, "	mytid = %d\n", mytid );
		response = NULL;
		assert( syscall_sock->code(response) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_pvm_info( response , mytid);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		free( (char *)response );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_pvm_task_info:
	  {
		int   v_pid;
		int   pvm_tid;
		condor_errno_t terrno;

		assert( syscall_sock->code(v_pid) );
		dprintf( D_SYSCALLS, "	v_pid = %d\n", v_pid );
		assert( syscall_sock->code(pvm_tid) );
		dprintf( D_SYSCALLS, "	pvm_tid = %d\n", pvm_tid );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_pvm_task_info( v_pid , pvm_tid);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_suspended:
	  {
		int   susp;
		condor_errno_t terrno;

		assert( syscall_sock->code(susp) );
		dprintf( D_SYSCALLS, "	susp = %d\n", susp );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_suspended( susp);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_get_file_stream:
	  {
		char *  file;
		size_t *  len;
		unsigned int *  ip_addr;
		u_short *  port_num;
		condor_errno_t terrno;

		file = NULL;
		port_num = (u_short *)malloc( (unsigned)U_SHORT_SIZE );
		memset( port_num, 0, (unsigned)U_SHORT_SIZE );
		ip_addr = (unsigned int *)malloc( (unsigned)U_INT_SIZE );
		memset( ip_addr, 0, (unsigned)U_INT_SIZE );
		len = (size_t *)malloc( (unsigned)SIZE_T_SIZE );
		memset( len, 0, (unsigned)SIZE_T_SIZE );
		assert( syscall_sock->code(file) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_get_file_stream( file , len , ip_addr , port_num);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		if( rval >= 0 ) {
			assert( syscall_sock->code(port_num) );
			assert( syscall_sock->code(ip_addr) );
			assert( syscall_sock->code(len) );
		}
		free( (char *)port_num );
		free( (char *)ip_addr );
		free( (char *)len );
		free( (char *)file );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_put_file_stream:
	  {
		char *  file;
		size_t   len;
		unsigned int *  ip_addr;
		u_short *  port_num;
		condor_errno_t terrno;

		assert( syscall_sock->code(len) );
		dprintf( D_SYSCALLS, "	len = %d\n", len );
		file = NULL;
		port_num = (u_short *)malloc( (unsigned)U_SHORT_SIZE );
		memset( port_num, 0, (unsigned)U_SHORT_SIZE );
		ip_addr = (unsigned int *)malloc( (unsigned)U_INT_SIZE );
		memset( ip_addr, 0, (unsigned)U_INT_SIZE );
		assert( syscall_sock->code(file) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_put_file_stream( file , len , ip_addr , port_num);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		if( rval >= 0 ) {
			assert( syscall_sock->code(port_num) );
			assert( syscall_sock->code(ip_addr) );
		}
		free( (char *)port_num );
		free( (char *)ip_addr );
		free( (char *)file );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_register_ckpt_server:
	  {
		char *  host;
		condor_errno_t terrno;

		host = NULL;
		assert( syscall_sock->code(host) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_register_ckpt_server( host);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		free( (char *)host );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_register_arch:
	  {
		char *  arch;
		condor_errno_t terrno;

		arch = NULL;
		assert( syscall_sock->code(arch) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_register_arch( arch);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		free( (char *)arch );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_register_opsys:
	  {
		char *  opsys;
		condor_errno_t terrno;

		opsys = NULL;
		assert( syscall_sock->code(opsys) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_register_opsys( opsys);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		free( (char *)opsys );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_register_begin_execution:
	  {
		condor_errno_t terrno;

		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_register_begin_execution( );
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_send_a_file:
	  {
		char *  name;
		mode_t   mode;
		condor_errno_t terrno;

		assert( syscall_sock->code(mode) );
		name = NULL;
		assert( syscall_sock->code(name) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_send_a_file( name , mode);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		free( (char *)name );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

	case CONDOR_get_file:
	  {
		char *  name;
		condor_errno_t terrno;

		name = NULL;
		assert( syscall_sock->code(name) );
		assert( syscall_sock->end_of_message() );;

		errno = (condor_errno_t)0;
		rval = pseudo_get_file( name);
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		if( rval < 0 ) {
			assert( syscall_sock->code(terrno) );
		}
		free( (char *)name );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}

/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

	default:
	{
		condor_errno_t terrno;

		/* consume arguments */
		syscall_sock->end_of_message();

		errno = 0;
		rval = CONDOR_NotSupported( condor_sysnum );
		terrno = (condor_errno_t)errno;
		dprintf( D_SYSCALLS, "\trval = %d, errno = %d\n", rval, (int)terrno );

		syscall_sock->encode();
		assert( syscall_sock->code(rval) );
		assert( syscall_sock->code(terrno) );
		assert( syscall_sock->end_of_message() );;
		return 0;
	}


}	/* End of switch on system call number */
}	/* End of do_REMOTE_syscall() procedure */

int
CONDOR_NotSupported( int condor_sysnum,  ... )
{
	job_report_store_error( "%s - system call not supported by Condor",
	                      _condor_syscall_name(condor_sysnum) );

	errno = ENOSYS;
	return -1;
}

int
CONDOR_Ignored( int condor_sysnum,  ... )
{
	dprintf(
		D_SYSCALLS,
		"%s - system call not supported by Condor - IGNORED!",
		_condor_syscall_name(condor_sysnum)
	);
	return 0;
}
